[{"locale":"es"},{"key":"b","names":["b"],"mappings":{"default":{"default":"bit"}}},{"key":"B","names":["B"],"mappings":{"default":{"default":"byte"}}},{"key":"KB","names":["KB"],"mappings":{"default":{"default":"kilobyte"}}},{"key":"MB","names":["MB"],"mappings":{"default":{"default":"megabyte"}}},{"key":"GB","names":["GB"],"mappings":{"default":{"default":"gigabyte"}}},{"key":"TB","names":["TB"],"mappings":{"default":{"default":"terabyte"}}}]
